// #define VERBOSE

/**
 * @file perf_toy_cube.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#if defined YAC_NETCDF_ENABLED
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "xml/config_xml.h"
#include "utils.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "yac_mpi.h"
#include "generate_cubed_sphere.h"
#include "test_function.h"

// #define VTK_OUTPUT
#ifdef VTK_OUTPUT
#include "vtk_output.h"
#endif

/* ------------------------------------------------- */

const char * fieldName[] = {"icon_out", "cube_out"};

static void generate_part_cube_grid_information(
  unsigned n, unsigned * nbr_vertices, unsigned * nbr_cells,
  unsigned ** num_vertices_per_cell, unsigned ** cell_to_vertex,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** global_cell_id, int ** global_cell_id_rank,
  int ** global_corner_id, int ** global_corner_id_rank, int rank, int size);

int main () {

  // Initialisation of MPI

  MPI_Init (NULL, NULL);

  xt_initialize(MPI_COMM_WORLD);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  // move xml and xsd file to node cache
  {
    FILE *f = fopen("perf_toy.xml", "rb");
    fseek(f, 0, SEEK_END);
    long fsize = ftell(f);
    fseek(f, 0, SEEK_SET);  /* same as rewind(f); */

    char *string = malloc(fsize + 1);
    fread(string, 1, fsize, f);
    fclose(f);
    free(string);
  }

  {
    FILE *f = fopen("coupling.xsd", "rb");
    fseek(f, 0, SEEK_END);
    long fsize = ftell(f);
    fseek(f, 0, SEEK_SET);  /* same as rewind(f); */

    char *string = malloc(fsize + 1);
    fread(string, 1, fsize, f);
    fclose(f);
    free(string);
  }

  yac_cinit ("perf_toy.xml", "coupling.xsd");

  /* The usual component definition, here for two sequential components on the
     same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "CUBE";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf(". main: defined %s with local comp ID %i \n", comp_name, comp_id);
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;

  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  yac_mpi_call(MPI_Comm_free(&local_comm), MPI_COMM_WORLD);

  int ierror = 0;

  int cell_point_id;
  int corner_point_id;

  int field_ids[2];
  int grid_id;

  unsigned n = 405;
  // unsigned n = 202;

  unsigned nbr_vertices;
  unsigned nbr_cells;
  unsigned * num_vertices_per_cell;
  unsigned * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * global_cell_id;
  int * global_cell_id_rank;
  int * global_corner_id;
  int * global_corner_id_rank;

  generate_part_cube_grid_information(n, &nbr_vertices, &nbr_cells,
                                      &num_vertices_per_cell, &cell_to_vertex,
                                      &x_vertices, &y_vertices, &x_cells,
                                      &y_cells, &global_cell_id,
                                      &global_cell_id_rank, &global_corner_id,
                                      &global_corner_id_rank, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  yac_cdef_grid_unstruct(
    "cube_grid", nbr_vertices, nbr_cells, (int*)num_vertices_per_cell,
    x_vertices, y_vertices, (int*)cell_to_vertex, &grid_id);

  int * cell_core_mask = xmalloc(nbr_cells * sizeof(*cell_core_mask));
  int * corner_core_mask = xmalloc(nbr_vertices * sizeof(*corner_core_mask));
  for (int i = 0; i < nbr_cells; ++i)
    cell_core_mask[i] = global_cell_id_rank[i] == -1;
  for (int i = 0; i < nbr_vertices; ++i)
    corner_core_mask[i] = global_corner_id_rank[i] == -1;

  yac_cset_global_index(global_cell_id, YAC_LOCATION_CELL, grid_id);
  yac_cset_core_mask(cell_core_mask, YAC_LOCATION_CELL, grid_id);
  yac_cset_global_index(global_corner_id, YAC_LOCATION_CORNER, grid_id);
  yac_cset_core_mask(corner_core_mask, YAC_LOCATION_CORNER, grid_id);

  yac_cdef_points_unstruct(
    grid_id, nbr_cells, YAC_LOCATION_CELL, x_center, y_center, &cell_point_id);
  yac_cdef_points_unstruct(
    grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_points, y_points, &corner_point_id);

  /* Field definition for the component */

  yac_cdef_field(
    fieldName[0], comp_id, &cell_point_id, 1, &field_ids[0]);
  yac_cdef_field(
    fieldName[1], comp_id, &cell_point_id, 1, &field_ids[1]);
  
  toc=MPI_Wtime();
  time = toc-tic;
  printf ("CUBE: Time for initialisation %f\n", time ); 

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch(&ierror);

  toc=MPI_Wtime();
  time = toc-tic;
  printf("CUBE: Time for search %f\n", time ); 

  double * conserv_in = xmalloc(nbr_cells * sizeof(*conserv_in));
  double * avg_in = xmalloc(nbr_vertices * sizeof(*avg_in));

  for (unsigned i = 0; i < nbr_cells; ++i)
    conserv_in[i] = -10;
  for (unsigned i = 0; i < nbr_vertices; ++i) avg_in[i] = -10;

  double * cell_out = xmalloc(nbr_cells * sizeof(*cell_out));
  double * vertex_out = xmalloc(nbr_vertices * sizeof(*vertex_out));

  int err;
  int info;

  for (unsigned i = 0; i < nbr_cells; ++i)
    cell_out[i] = test_func(x_center[i], y_center[i]);
  for (unsigned i = 0; i < nbr_vertices; ++i)
    vertex_out[i] = test_func(x_vertices[i], y_vertices[i]);

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  {
    double *point_set_data[1];
    double **collection_data[1] = {point_set_data};

    point_set_data[0] = cell_out;
    yac_cput(field_ids[1], 1, collection_data, &info, &err);
  }

  {
    double *collection_data[1] = {conserv_in};

    yac_cget(field_ids[0], 1, collection_data, &info, &err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("CUBE: Time for ping-pong %f\n", time );

#ifdef VTK_OUTPUT
  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "perf_toy_cube_%d.vtk", rank);

  double point_data[nbr_vertices][3];
  for (int i = 0; i < nbr_vertices; ++i) {
   LLtoXYZ(x_vertices[i], y_vertices[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "cube_out");
  vtk_write_point_data(vtk_file, (double *)point_data, nbr_vertices);
  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, nbr_cells);
  vtk_write_point_scalars_int(
    vtk_file, corner_core_mask, nbr_vertices, "corner_core_mask");
  vtk_write_point_scalars_int(
    vtk_file, global_corner_id, nbr_vertices, "global_corner_id");
  vtk_write_cell_scalars_int(
    vtk_file, cell_core_mask, nbr_cells, "cell_core_mask");
  vtk_write_cell_scalars_int(
    vtk_file, global_cell_id, nbr_cells, "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in, nbr_cells, "conserv_in");
  vtk_write_cell_scalars_double(vtk_file, cell_out, nbr_cells, "cell_out");
  vtk_write_point_scalars_double(vtk_file, avg_in, nbr_vertices, "avg_in");
  vtk_write_point_scalars_double(vtk_file, vertex_out, nbr_vertices, "vertex_out");

  vtk_close(vtk_file);
#endif // VTK_OUTPUT

  free(corner_core_mask);
  free(global_corner_id_rank);
  free(global_corner_id);
  free(cell_core_mask);
  free(global_cell_id_rank);
  free(global_cell_id);
  free(x_cells);
  free(y_cells);
  free(x_vertices);
  free(y_vertices);
  free(num_vertices_per_cell);
  free(cell_to_vertex);

  yac_cfinalize();

  xt_finalize();

  MPI_Finalize();

  free(conserv_in);
  free(cell_out);
  free(avg_in);
  free(vertex_out);

  return EXIT_SUCCESS;
}

static void decompose_domain_simple(unsigned n, int size, int * cell_owner) {

  unsigned nbr_cells = n * n * 6;

  for (unsigned i = 0; i < nbr_cells; ++i)
    cell_owner[i] = (i * size + size - 1) / nbr_cells;
}

static void compute_reg2d_decomp(int size, int * num_procs) {

  int best_num_procs[2] = {size, 1};
  float curr_proc_ratio = (float)size;
  float best_proc_ratio = (float)size;

  for (int i = size - 1; i > 0; --i) {

    if (size % i == 0) {

      curr_proc_ratio = (float)(i) / (float)(size/i);

      double temp[2] =
        {fabs(1.0 - curr_proc_ratio), fabs(1.0 - best_proc_ratio)};
      if (temp[0] < temp[1]) {
        best_num_procs[0] = i;
        best_num_procs[1] = size/i;
        best_proc_ratio = curr_proc_ratio;
      }
    }
  }

  num_procs[0] = best_num_procs[0];
  num_procs[1] = best_num_procs[1];
}

static void decompose_domain_2d(unsigned n, int size, int * cell_owner) {

  // distribute processes among all six sides of the cubed sphere
  int proc_start[6 + 1];
  for (int i = 0; i < 6 + 1; ++i)
    proc_start[i] = (size * i) / 6;

  // for each side of the cube
  for (unsigned i = 0, offset = 0; i < 6; ++i) {

    int num_procs[2];
    int curr_num_procs = proc_start[i+1] - proc_start[i];

    compute_reg2d_decomp(curr_num_procs, num_procs);

    int local_start_x[num_procs[0] + 1];
    int local_start_y[num_procs[1] + 1];

    for (int j = 0; j < num_procs[0] + 1; ++j)
      local_start_x[j] = (n * j) / num_procs[0];
    for (int j = 0; j < num_procs[1] + 1; ++j)
      local_start_y[j] = (n * j) / num_procs[1];

    for (int p_y = 0; p_y < num_procs[1]; ++p_y) {
      for (int row = local_start_y[p_y]; row < local_start_y[p_y+1]; ++row) {
        for (int p_x = 0; p_x < num_procs[0]; ++p_x) {
          for (int column = local_start_x[p_x]; column < local_start_x[p_x+1];
               ++column) {

            cell_owner[offset++] = proc_start[i] + p_x + num_procs[0] * p_y;
          }
        }
      }
    }
  }
}

static void decompose_domain(unsigned n, int size, int * cell_owner) {

  if (size <= 6) {
    decompose_domain_simple(n, size, cell_owner);
  } else {
    decompose_domain_2d(n, size, cell_owner);
  }
}

static void generate_part_cube_grid_information(
  unsigned n, unsigned * nbr_vertices, unsigned * nbr_cells,
  unsigned ** num_vertices_per_cell, unsigned ** cell_to_vertex,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** global_cell_id, int ** global_cell_id_rank,
  int ** global_corner_id, int ** global_corner_id_rank, int rank, int size) {

  double * x_vertices_3d, * y_vertices_3d, * z_vertices_3d;
  unsigned * dummy;

  // generate global grid
  generate_cubed_sphere_grid_information(n, nbr_cells, nbr_vertices,
                                         &x_vertices_3d, &y_vertices_3d,
                                         &z_vertices_3d, cell_to_vertex,
                                         &dummy);
  *num_vertices_per_cell = xmalloc(*nbr_cells * sizeof(**num_vertices_per_cell));
  for (unsigned i = 0; i < *nbr_cells; ++i) (*num_vertices_per_cell)[i] = 4;
  free(dummy);

  double * x_cell_3d = xmalloc(*nbr_cells * sizeof(*x_cell_3d));
  double * y_cell_3d = xmalloc(*nbr_cells * sizeof(*y_cell_3d));
  double * z_cell_3d = xmalloc(*nbr_cells * sizeof(*z_cell_3d));

  for (unsigned i = 0; i < *nbr_cells; ++i) {

    x_cell_3d[i] = y_cell_3d[i] = z_cell_3d[i] = 0;

    for (unsigned j = 0; j < 4; ++j) {

      x_cell_3d[i] += x_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      y_cell_3d[i] += y_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      z_cell_3d[i] += z_vertices_3d[(*cell_to_vertex)[4 * i + j]];
    }

    double scale = 1.0 / sqrt(x_cell_3d[i] * x_cell_3d[i] +
                              y_cell_3d[i] * y_cell_3d[i] +
                              z_cell_3d[i] * z_cell_3d[i]);

    x_cell_3d[i] *= scale;
    y_cell_3d[i] *= scale;
    z_cell_3d[i] *= scale;
  }

  int * cell_is_on_rank = xmalloc(*nbr_cells * sizeof(*cell_is_on_rank));

  decompose_domain(n, size, cell_is_on_rank);

  // mask for required vertices and cells
  int * required_vertices = xcalloc(*nbr_vertices, sizeof(*required_vertices));
  int * required_cells = xcalloc(*nbr_cells, sizeof(*required_cells));

  // mark all local cells and their vertices as required
  for (unsigned i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (cell_is_on_rank[i] != rank) continue;

    cell_is_on_rank[i] = -1;

    required_cells[i] = 1;
    for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;
  }

  // mark all halo cells as required and generate cell_is_on_rank
  for (unsigned i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (!required_cells[i]) {

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          break;
        }
      }

    }
  }

  // mask for halo vertices
  int * vertex_is_on_rank = xmalloc(*nbr_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  for (unsigned i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (unsigned i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (unsigned i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (unsigned i = 0; i < *nbr_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = xmalloc(part_num_cells * sizeof(**global_cell_id));
  *global_cell_id_rank = xmalloc(part_num_cells * sizeof(**global_cell_id_rank));
  *global_corner_id = xmalloc(part_num_vertices * sizeof(**global_corner_id));
  *global_corner_id_rank = xmalloc(part_num_vertices * sizeof(**global_corner_id_rank));

  *x_cells = xmalloc(part_num_cells * sizeof(**x_cells));
  *y_cells = xmalloc(part_num_cells * sizeof(**y_cells));
  *x_vertices = xmalloc(part_num_vertices * sizeof(**x_vertices));
  *y_vertices = xmalloc(part_num_vertices * sizeof(**y_vertices));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = xmalloc(*nbr_vertices * sizeof(*global_to_local_vertex));
  for (unsigned i = 0; i < *nbr_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*global_corner_id_rank)[part_num_vertices] = vertex_is_on_rank[i];
      double p[3] = {x_vertices_3d[i], y_vertices_3d[i], z_vertices_3d[i]};
      XYZtoLL(p, (*x_vertices) + part_num_vertices, (*y_vertices) + part_num_vertices);
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  free(vertex_is_on_rank);
  *nbr_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  for (unsigned i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*global_cell_id_rank)[part_num_cells] = cell_is_on_rank[i];
      double middle_point[3] = {x_cell_3d[i],y_cell_3d[i],z_cell_3d[i]};

      

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {
        (*cell_to_vertex)[num_cell_vertex_dependencies+j] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];
      }

      XYZtoLL(middle_point, (*x_cells)+part_num_cells, (*y_cells)+part_num_cells);

      num_cell_vertex_dependencies += (*num_vertices_per_cell)[i];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }
  }

  free(x_cell_3d);
  free(y_cell_3d);
  free(z_cell_3d);
  free(x_vertices_3d);
  free(y_vertices_3d);
  free(z_vertices_3d);

  *num_vertices_per_cell = xrealloc(*num_vertices_per_cell, part_num_cells *
                                    sizeof(**num_vertices_per_cell));
  *cell_to_vertex = xrealloc(*cell_to_vertex, num_cell_vertex_dependencies *
                             sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *nbr_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with NetCDF.\n");
  return EXIT_FAILURE;
}
#endif

