/**
 * @file test_dummy_coupling3.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "yac_config.h"

#define EXACT

#include <mpi.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_interface.h"
#include "read_icon_grid.h"
#include "read_mpiom_grid.h"
#include "generate_cubed_sphere.h"
#include "test_function.h"
#include "utils.h"

#include "tests.h"

#define DUMMY_VALUE (-1337.0)

struct grid_data {

  char const * name;

  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * global_cell_id;
  int * cell_core_mask;
  int * global_corner_id;
  int * corner_core_mask;
};

struct intra_comp_yac_info {
  int yac_id;
  int is_comp_root;
  int nbr_vertices;
  int out_field_id;
  int in_field_id;
};

struct comp_pair_yac_info {
  int yac_id;
  int nbr_cells[2];
  int out_field_ids[2];
  int in_field_ids[2];
};

static struct intra_comp_yac_info setup_intra_comp_yac(
  MPI_Comm comp_comm, struct grid_data grid_data);

static struct comp_pair_yac_info setup_comp_pair_yac(
  char const * comp_pair_names[2], struct grid_data comp_grid_data[2],
  int local_comp_indices[2]);

static struct grid_data generate_icon_grid_data(MPI_Comm comm);
static struct grid_data generate_cube_grid_data(MPI_Comm comm);
static struct grid_data generate_mpiom_grid_data(MPI_Comm comm);
static void grid_data_free(struct grid_data grid_data);

static void def_default_fields(
  int comp_id, char const * comp_name, int grid_id, int cell_point_id,
  int * in_field_ids, int * out_field_id);

/**
 * This tests checks the support for multiple parallel YAC instances.
 * There are three main components and one dummy component. The dummy component
 * does not take part in any coupling.
 *
 * All four components are registered in the default YAC instance.
 *
 * The three active components couple to each other.
 *
 * Each active component shares one process with each of the two other active
 * components.
 *
 * Each active component pair generates their own YAC instance.
 *
 * Each active component has its own YAC instance in which one of its processes
 * generates its own component.
 *
 * The whole setup is run twice in order to simulate the restarting of YAC
 * instances.
 */
int main (void) {
#ifdef YAC_NETCDF_ENABLED

  MPI_Init(NULL, NULL);
  xt_initialize(MPI_COMM_WORLD);

  int global_rank, global_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &global_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &global_size);

  if (global_size != 8) {
    fprintf(stderr, "Wrong number of processes (should be 8)\n");
    exit(EXIT_FAILURE);
  }

  // determine the local components
  char const * global_comp_names[] = {"comp_a", "comp_b", "comp_c", "dummy"};
  char const * local_comp_names[2];
  int num_local_comps = 0;
  int is_dummy = ((global_rank == 3) || (global_rank == 7));
  if ((global_rank == 0) || (global_rank == 4) || (global_rank == 5))
    local_comp_names[num_local_comps++] = global_comp_names[0];
  if ((global_rank == 1) || (global_rank == 4) || (global_rank == 6))
    local_comp_names[num_local_comps++] = global_comp_names[1];
  if ((global_rank == 2) || (global_rank == 5) || (global_rank == 6))
    local_comp_names[num_local_comps++] = global_comp_names[2];
  if ((global_rank == 3) || (global_rank == 7))
    local_comp_names[num_local_comps++] = global_comp_names[3];

  // initialise default YAC instance
  yac_cinit ("coupling_test3_default.xml", "coupling.xsd");

  // register local components in default YAC instance
  int default_comp_ids[2];
  int default_num_comps = num_local_comps;
  yac_cdef_comps(local_comp_names, default_num_comps, default_comp_ids);

  MPI_Comm default_comp_comms[2];
  for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
    yac_cget_localcomm(
      &(default_comp_comms[comp_idx]), default_comp_ids[comp_idx]);

  // generate grid_data
  struct grid_data comp_grid_data[2];
  for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
    switch(local_comp_names[comp_idx][5]) {
      case('a'): {
        comp_grid_data[comp_idx] =
         generate_icon_grid_data(default_comp_comms[comp_idx]);
        break;
      }
      case('b'): {
        comp_grid_data[comp_idx] =
          generate_cube_grid_data(default_comp_comms[comp_idx]);
        break;
      }
      case('c'): {
        comp_grid_data[comp_idx] =
          generate_mpiom_grid_data(default_comp_comms[comp_idx]);
        break;
      }
      default:
        break;
    };
  }

  // setup yac instance which couples a component to one of its own ranks
  struct intra_comp_yac_info yac_intra_infos[2];
  if (!is_dummy)
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
      yac_intra_infos[comp_idx] =
        setup_intra_comp_yac(
          default_comp_comms[comp_idx], comp_grid_data[comp_idx]);

  // register grid, global ids and core mask
  int default_grid_ids[2];
  if (!is_dummy) {
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
      yac_cdef_grid_unstruct(
        comp_grid_data[comp_idx].name,
        comp_grid_data[comp_idx].nbr_vertices,
        comp_grid_data[comp_idx].nbr_cells,
        comp_grid_data[comp_idx].num_vertices_per_cell,
        comp_grid_data[comp_idx].x_vertices,
        comp_grid_data[comp_idx].y_vertices,
        comp_grid_data[comp_idx].cell_to_vertex,
        &(default_grid_ids[comp_idx]));
      yac_cset_global_index(
        comp_grid_data[comp_idx].global_corner_id, YAC_LOCATION_CORNER,
        default_grid_ids[comp_idx]);
      yac_cset_core_mask(
        comp_grid_data[comp_idx].corner_core_mask, YAC_LOCATION_CORNER,
        default_grid_ids[comp_idx]);
      yac_cset_global_index(
        comp_grid_data[comp_idx].global_cell_id, YAC_LOCATION_CELL,
        default_grid_ids[comp_idx]);
      yac_cset_core_mask(
        comp_grid_data[comp_idx].cell_core_mask, YAC_LOCATION_CELL,
        default_grid_ids[comp_idx]);
    }
  }

  // register cell points and a default mask if available
  int cell_point_ids[2];
  if (!is_dummy) {
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
      yac_cdef_points_unstruct(
        default_grid_ids[comp_idx], comp_grid_data[comp_idx].nbr_cells,
        YAC_LOCATION_CELL, comp_grid_data[comp_idx].x_cells,
        comp_grid_data[comp_idx].y_cells, &(cell_point_ids[comp_idx]));
      if (comp_grid_data[comp_idx].cell_mask != NULL)
        yac_cset_mask(
          comp_grid_data[comp_idx].cell_mask, cell_point_ids[comp_idx]);
    }
  }
  
  // register in and out fields
  int default_in_field_ids[2][2];// = {{INT_MAX, INT_MAX}, {INT_MAX, INT_MAX}};
  int default_out_field_ids[2];// = {INT_MAX, INT_MAX};
  if (!is_dummy)
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
      def_default_fields(
        default_comp_ids[comp_idx], local_comp_names[comp_idx],
        default_grid_ids[comp_idx], cell_point_ids[comp_idx],
        default_in_field_ids[comp_idx], &(default_out_field_ids[comp_idx]));

  // generate interpolations for the default YAC instance
  int ierror;
  yac_csearch(&ierror);

  // setup yac instances that only couple all component combinations
  struct comp_pair_yac_info yac_comp_pair_infos[3];
  for (int i = 0, k = 0; i < 3; ++i) {

    int local_comp_idx[2];

    local_comp_idx[0] = INT_MAX;
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
      if (!strcmp(global_comp_names[i], local_comp_names[comp_idx]))
        local_comp_idx[0] = comp_idx;

    for (int j = i + 1; j < 3; ++j, ++k) {

      local_comp_idx[1] = INT_MAX;
      for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
        if (!strcmp(global_comp_names[j], local_comp_names[comp_idx]))
          local_comp_idx[1] = comp_idx;

      char const * comp_pair_names[2] =
        {global_comp_names[i], global_comp_names[j]};

      yac_comp_pair_infos[k] =
        setup_comp_pair_yac(
          comp_pair_names, comp_grid_data, local_comp_idx);
    }
  }

  // if this is not the dummy process, which is not involved in any coupling
  if (!is_dummy) {

    double * out_field_data[2];
    double * in_field_data[2];
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
      out_field_data[comp_idx] =
        xmalloc(comp_grid_data[comp_idx].nbr_cells * sizeof(double));
      for (int i = 0; i < comp_grid_data[comp_idx].nbr_cells; ++i)
        out_field_data[comp_idx][i] =
          (comp_grid_data[comp_idx].cell_core_mask[i])?
            (test_harmonic(
               comp_grid_data[comp_idx].x_cells[i],
               comp_grid_data[comp_idx].y_cells[i])):DUMMY_VALUE;
      in_field_data[comp_idx] =
        xmalloc(comp_grid_data[comp_idx].nbr_cells * sizeof(double));
    }

    // do some ping-pongs in the default yac instance
    for (int t = 0; t < 10; ++t) {

      for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
        int info, err;
        int id = default_out_field_ids[comp_idx];
        double *point_set_data[1];
        double **collection_data[1] = {point_set_data};
        point_set_data[0] = out_field_data[comp_idx];
        yac_cput(id, 1, collection_data, &info, &err);
      }

      for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
        for (int field_idx = 0; field_idx < 2; ++field_idx) {
          int info, err;
          int id = default_in_field_ids[comp_idx][field_idx];
          double *collection_data[1] = {in_field_data[comp_idx]};

          for (int i = 0; i < comp_grid_data[comp_idx].nbr_cells; ++i)
            in_field_data[comp_idx][i] = DUMMY_VALUE;
          yac_cget(id, 1, collection_data, &info, &err);
        }
      }
    }

    // interpolate received results to the regional grid defined on
    // rank 0 of the current component
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
      
      if (yac_intra_infos[comp_idx].is_comp_root) {
        int out_info, in_info, err;
        int out_id = yac_intra_infos[comp_idx].out_field_id;
        double *out_point_set_data[1];
        double **out_collection_data[1] = {out_point_set_data};
        out_point_set_data[0] = in_field_data[comp_idx];

        int in_id = yac_intra_infos[comp_idx].in_field_id;
        double *in_collection_data[1] = {
          xmalloc(yac_intra_infos[comp_idx].nbr_vertices * sizeof(double))};
        for (int i = 0; i < yac_intra_infos[comp_idx].nbr_vertices; ++i)
          in_collection_data[0][i] = DUMMY_VALUE;

        yac_cexchange(
          out_id, in_id, 1, out_collection_data, in_collection_data,
          &out_info, &in_info, &err);

        free(in_collection_data[0]);
      } else {
        int info, err;
        int id = yac_intra_infos[comp_idx].out_field_id;
        double *point_set_data[1];
        double **collection_data[1] = {point_set_data};
        point_set_data[0] = in_field_data[comp_idx];
        yac_cput(id, 1, collection_data, &info, &err);
      }
    }

    // do some exchanges between within the component pair YAC instances
    // for all available component pairs
    for (int i = 0; i < 3; ++i) {

      // if the local process is not a part of the current component pair
      if (yac_comp_pair_infos[i].yac_id == INT_MAX) continue;

      // if the local process has both components of the current pair
      if ((yac_comp_pair_infos[i].nbr_cells[0] > 0) &&
          (yac_comp_pair_infos[i].nbr_cells[1] > 0)) {

        double * out_field_data[2];
        double * in_field_data[2];

        for (int j = 0; j < 2; ++j) {
          int nbr_cells = yac_comp_pair_infos[i].nbr_cells[j];
          out_field_data[j] = xmalloc(nbr_cells * sizeof(double));
          for (int k = 0; k < nbr_cells; ++k) out_field_data[j][k] = 0.0;
          in_field_data[j] = xmalloc(nbr_cells * sizeof(double));
          for (int k = 0; k < nbr_cells; ++k) in_field_data[j][k] = DUMMY_VALUE;
        }

        for (int j = 0; j < 2; ++j) {

          int out_info, in_info, err;
          int out_id = yac_comp_pair_infos[i].out_field_ids[j];
          double *out_point_set_data[1];
          double **out_collection_data[1] = {out_point_set_data};
          out_point_set_data[0] = out_field_data[j];

          int in_id = yac_comp_pair_infos[i].in_field_ids[j^1];
          double *in_collection_data[1] = {in_field_data[j^1]};

          yac_cexchange(
            out_id, in_id, 1, out_collection_data, in_collection_data,
            &out_info, &in_info, &err);
        }
        for (int j = 0; j < 2; ++j) {
          free(out_field_data[j]);
          free(in_field_data[j]);
        }

      } else {

        int flag = yac_comp_pair_infos[i].nbr_cells[1] > 0;
        int nbr_cells = yac_comp_pair_infos[i].nbr_cells[flag];

        {

          double * out_field_data = xmalloc(nbr_cells * sizeof(double));
          // just some dummy data
          for (int j = 0; j < nbr_cells; ++j) out_field_data[j] = 0.0;

          int info, err;
          int id = yac_comp_pair_infos[i].out_field_ids[flag];
          double *point_set_data[1];
          double **collection_data[1] = {point_set_data};
          point_set_data[0] = out_field_data;
          yac_cput(id, 1, collection_data, &info, &err);

          free(out_field_data);
        }
        {

          double * in_field_data = xmalloc(nbr_cells * sizeof(double));

          int info, err;
          int id = yac_comp_pair_infos[i].in_field_ids[flag];
          double *collection_data[1] = {in_field_data};

          for (int j = 0; j < nbr_cells; ++j) in_field_data[j] = DUMMY_VALUE;
          yac_cget(id, 1, collection_data, &info, &err);

          free(in_field_data);
        }
      }
    }

    // clean-up

    for (size_t comp_idx = 0; comp_idx < num_local_comps; ++comp_idx) {
      free(out_field_data[comp_idx]);
      free(in_field_data[comp_idx]);
      grid_data_free(comp_grid_data[comp_idx]);
    }
  }

  for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
    MPI_Comm_free(&(default_comp_comms[comp_idx]));

  if (!is_dummy) {
    for (int i = 0; i < 3; ++i)
      if (yac_comp_pair_infos[i].yac_id != INT_MAX)
        yac_ccleanup_instance(yac_comp_pair_infos[i].yac_id);
    for (int comp_idx = 0; comp_idx < num_local_comps; ++comp_idx)
      yac_ccleanup_instance(yac_intra_infos[comp_idx].yac_id);
  }
  yac_cfinalize();

  xt_finalize();
  MPI_Finalize();

#else

  PUT_ERR("YAC configured without XML or NetCDF\n");

#endif

  return TEST_EXIT_CODE;
}

static struct intra_comp_yac_info setup_intra_comp_yac(
  MPI_Comm comp_comm, struct grid_data base_grid_data) {

  int comp_rank, comp_size;
  MPI_Comm_rank(comp_comm, &comp_rank);
  MPI_Comm_size(comp_comm, &comp_size);

  // initialise intra component YAC instance
  int yac_id;
  yac_cinit_comm_instance(
    "coupling_test3_intra.xml", "coupling.xsd", comp_comm, &yac_id);

  // register component(s) (all processes are part of the base component
  //                        process 0 has an additional component)
  char const * comp_names[2] = {"comp_base", "comp_regional"};
  int num_comps = (comp_rank == 0)?2:1;
  int comp_ids[2];
  yac_cdef_comps_instance(yac_id, comp_names, num_comps, comp_ids);

  // register grid, global ids and core mask for base component
  int grid_ids[2];
  yac_cdef_grid_unstruct_instance(
    yac_id, "base_grid", base_grid_data.nbr_vertices, base_grid_data.nbr_cells,
    base_grid_data.num_vertices_per_cell,
    base_grid_data.x_vertices, base_grid_data.y_vertices,
    base_grid_data.cell_to_vertex, &(grid_ids[0]));
  yac_cset_global_index(
    base_grid_data.global_corner_id, YAC_LOCATION_CORNER, grid_ids[0]);
  yac_cset_core_mask(
    base_grid_data.corner_core_mask, YAC_LOCATION_CORNER, grid_ids[0]);
  yac_cset_global_index(
    base_grid_data.global_cell_id, YAC_LOCATION_CELL, grid_ids[0]);
  yac_cset_core_mask(
    base_grid_data.cell_core_mask, YAC_LOCATION_CELL, grid_ids[0]);


  // register regional grid on rank 0
  double regional_coord_x[] = {
    -20,-18,-16,-14,-12,-10, -8, -6, -4, -2,
      0,  2,  4,  6,  8, 10, 12, 14, 16, 18, 20};
  double regional_coord_y[] = {
    -20,-18,-16,-14,-12,-10, -8, -6, -4, -2,
      0,  2,  4,  6,  8, 10, 12, 14, 16, 18, 20};
  int num_vertices_regional[2] =
    {sizeof(regional_coord_x) / sizeof(regional_coord_x[0]),
     sizeof(regional_coord_y) / sizeof(regional_coord_y[0])};
  if (comp_rank == 0) {
    int cyclic[2] = {0, 0};
    for (int i = 0; i < num_vertices_regional[0]; ++i)
      regional_coord_x[i] *= YAC_RAD;
    for (int i = 0; i < num_vertices_regional[1]; ++i)
      regional_coord_y[i] *= YAC_RAD;
    yac_cdef_grid_reg2d_instance(
      yac_id, "regional_grid", num_vertices_regional, cyclic,
      regional_coord_x, regional_coord_y, &(grid_ids[1]));
  }

  // register cell points and a default mask if available
  int base_point_id, regional_point_id;
  yac_cdef_points_unstruct(
    grid_ids[0], base_grid_data.nbr_cells, YAC_LOCATION_CELL,
    base_grid_data.x_cells, base_grid_data.y_cells, &base_point_id);
  if (base_grid_data.cell_mask != NULL)
    yac_cset_mask(base_grid_data.cell_mask, base_point_id);
  if (comp_rank == 0) {
    yac_cdef_points_reg2d(
      grid_ids[1], num_vertices_regional, YAC_LOCATION_CORNER,
      regional_coord_x, regional_coord_y,  &regional_point_id);
  }
  
  // register field (from base component to regional one)
  int field_ids[2];
  yac_cdef_field(
    "base_to_regional", comp_ids[0], &base_point_id, 1, &(field_ids[0]));
  if (comp_rank == 0)
    yac_cdef_field(
      "base_to_regional", comp_ids[1], &regional_point_id, 1, &(field_ids[1]));

  // generate interpolations for the default YAC instance
  int ierror;
  yac_csearch_instance(yac_id, &ierror);

  struct intra_comp_yac_info yac_intra_info;
  yac_intra_info.yac_id = yac_id;
  yac_intra_info.is_comp_root = comp_rank == 0;
  yac_intra_info.nbr_vertices =
    num_vertices_regional[0] * num_vertices_regional[1];
  yac_intra_info.out_field_id = field_ids[0];
  yac_intra_info.in_field_id = field_ids[1];

  return yac_intra_info;
}

static struct comp_pair_yac_info setup_comp_pair_yac(
  char const * comp_pair_names[2], struct grid_data comp_grid_data[2],
  int local_comp_indices[2]) {

  if ((local_comp_indices[0] == INT_MAX) &&
      (local_comp_indices[1] == INT_MAX))
    return (struct comp_pair_yac_info){.yac_id = INT_MAX};

  // get the communicator for the component pair
  MPI_Comm comp_pair_comm;
  yac_cget_pair_rootcomm(
    &comp_pair_comm, comp_pair_names[0], comp_pair_names[1]);

  char xml_file_name[256];
  sprintf(xml_file_name, "coupling_test3_%s_%s.xml",
         comp_pair_names[0], comp_pair_names[1]);

  // initialise component pair YAC instance
  int yac_id;
  yac_cinit_comm_instance(
    xml_file_name, "coupling.xsd", comp_pair_comm, &yac_id);

  // register component(s) the local process can have one or two components
  // associated to the YAC instance
  char const * comp_names[2];
  int num_comps = 0;
  int comp_ids[2];
  for (int i = 0; i < 2; ++i)
    if (local_comp_indices[i] != INT_MAX)
      comp_names[num_comps++] = comp_pair_names[i];
  yac_cdef_comps_instance(yac_id, comp_names, num_comps, comp_ids);

  int grid_ids[2];
  int field_ids[2][2];
  for (int i = 0, k = 0; i < 2; ++i) {

    int comp_idx = local_comp_indices[i];
    if (comp_idx == INT_MAX) continue;

    struct grid_data curr_comp_grid_data = comp_grid_data[comp_idx];

    // register grid, global ids and core mask for the component(s)
    yac_cdef_grid_unstruct_instance(
      yac_id, curr_comp_grid_data.name, curr_comp_grid_data.nbr_vertices,
      curr_comp_grid_data.nbr_cells, curr_comp_grid_data.num_vertices_per_cell,
      curr_comp_grid_data.x_vertices, curr_comp_grid_data.y_vertices,
      curr_comp_grid_data.cell_to_vertex, &(grid_ids[i]));
    yac_cset_global_index(
      curr_comp_grid_data.global_corner_id, YAC_LOCATION_CORNER, grid_ids[i]);
    yac_cset_core_mask(
      curr_comp_grid_data.corner_core_mask, YAC_LOCATION_CORNER, grid_ids[i]);
    yac_cset_global_index(
      curr_comp_grid_data.global_cell_id, YAC_LOCATION_CELL, grid_ids[i]);
    yac_cset_core_mask(
      curr_comp_grid_data.cell_core_mask, YAC_LOCATION_CELL, grid_ids[i]);

    // register cell points and a default mask if available
    int point_id;
    yac_cdef_points_unstruct(
      grid_ids[i], curr_comp_grid_data.nbr_cells, YAC_LOCATION_CELL,
      curr_comp_grid_data.x_cells, curr_comp_grid_data.y_cells, &point_id);
    if (curr_comp_grid_data.cell_mask != NULL)
      yac_cset_mask(curr_comp_grid_data.cell_mask, point_id);

    // register field
    for (int j = 0; j < 2; ++j) {
      char field_name[16];
      sprintf(field_name, "comp_%c_out", comp_pair_names[j][5]);
      yac_cdef_field(field_name, comp_ids[k], &point_id, 1, &(field_ids[i][j]));
    }
    ++k;
  }

  // generate interpolations for the component pair YAC instance
  int ierror;
  yac_csearch_instance(yac_id, &ierror);

  MPI_Comm_free(&comp_pair_comm);

  struct comp_pair_yac_info yac_info;
  
  yac_info.yac_id = yac_id;
  for (int i = 0; i < 2; ++i) {
    int comp_idx = local_comp_indices[i];
    if (comp_idx == INT_MAX) {
      yac_info.nbr_cells[i] = 0;
      yac_info.out_field_ids[i] = INT_MAX;
      yac_info.in_field_ids[i] = INT_MAX;
    } else {
      yac_info.nbr_cells[i] = comp_grid_data[comp_idx].nbr_cells;
      int roles[2];
      for (int j = 0; j < 2; ++j) {
        yac_cget_role_of_field_by_id(field_ids[i][j], &(roles[j]));
        switch (roles[j]) {
          case(0):
          default:
            PUT_ERR("invalid role");
            break;
          case(1):
            yac_info.out_field_ids[i] = field_ids[i][j];
            break;
          case(2):
            yac_info.in_field_ids[i] = field_ids[i][j];
            break;
        }
      }
      if (roles[0] == roles[1]) PUT_ERR("roles should not match");
    }
  }

  return yac_info;
}

static void def_default_fields(
  int comp_id, char const * comp_name, int grid_id, int cell_point_id,
  int * in_field_ids, int * out_field_id) {

  int num_in_fields = 0;

  for (int i = 0; i < 3; ++i) {
    char const * field_names[3] = {"comp_a_out", "comp_b_out", "comp_c_out"};
    int field_id;
    char const * curr_field_name = field_names[i];
    yac_cdef_field(curr_field_name, comp_id, &cell_point_id, 1, &field_id);

    if (comp_name[5] == curr_field_name[5])
      *out_field_id = field_id;
    else
      in_field_ids[num_in_fields++] = field_id;
  }
}

static void generate_part_cube_grid_information(
  unsigned n, unsigned * nbr_vertices, unsigned * nbr_cells,
  unsigned ** num_vertices_per_cell, unsigned ** cell_to_vertex,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** global_cell_id, int ** cell_core_mask,
  int ** global_corner_id, int ** corner_core_mask, int rank, int size) {

  double * x_vertices_3d, * y_vertices_3d, * z_vertices_3d;
  unsigned * dummy;

  // generate global grid
  generate_cubed_sphere_grid_information(n, nbr_cells, nbr_vertices,
                                         &x_vertices_3d, &y_vertices_3d,
                                         &z_vertices_3d, cell_to_vertex,
                                         &dummy);
  *num_vertices_per_cell = xmalloc(*nbr_cells * sizeof(**num_vertices_per_cell));
  for (unsigned i = 0; i < *nbr_cells; ++i) (*num_vertices_per_cell)[i] = 4;
  free(dummy);

  double * x_cell_3d = xmalloc(*nbr_cells * sizeof(*x_cell_3d));
  double * y_cell_3d = xmalloc(*nbr_cells * sizeof(*y_cell_3d));
  double * z_cell_3d = xmalloc(*nbr_cells * sizeof(*z_cell_3d));

  for (unsigned i = 0; i < *nbr_cells; ++i) {

    x_cell_3d[i] = y_cell_3d[i] = z_cell_3d[i] = 0;

    for (unsigned j = 0; j < 4; ++j) {

      x_cell_3d[i] += x_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      y_cell_3d[i] += y_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      z_cell_3d[i] += z_vertices_3d[(*cell_to_vertex)[4 * i + j]];
    }

    double scale = 1.0 / sqrt(x_cell_3d[i] * x_cell_3d[i] + 
                              y_cell_3d[i] * y_cell_3d[i] + 
                              z_cell_3d[i] * z_cell_3d[i]);

    x_cell_3d[i] *= scale;
    y_cell_3d[i] *= scale;
    z_cell_3d[i] *= scale;
  }

  // determine local rank
  int num_cells_per_process = (*nbr_cells + size - 1)/size;
  int local_start = rank * num_cells_per_process;
  int num_local_cells =
    MIN(num_cells_per_process, (int)(*nbr_cells) - local_start);

  // mask for required vertices and cells
  int * required_vertices = xcalloc(*nbr_vertices, sizeof(*required_vertices));
  int * required_cells = xcalloc(*nbr_cells, sizeof(*required_cells));

  int offset = 0;
  for (int i = 0; i < local_start; ++i)
    offset += (*num_vertices_per_cell)[i];

  // mark all local cells and their vertices as required
  for (int i = local_start; i < local_start + num_local_cells; ++i) {

    required_cells[i] = 1;
    for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo cells
  int * cell_is_on_rank = xmalloc(*nbr_cells * sizeof(*cell_is_on_rank));

  // mark all halo cells as required and generate cell_is_on_rank
  offset = 0;
  for (unsigned i = 0; i < *nbr_cells; ++i) {

    if (!required_cells[i]) {

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          cell_is_on_rank[i] = i / num_cells_per_process;
          break;
        }
      }

    } else {
      cell_is_on_rank[i] = -1;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo vertices
  int * vertex_is_on_rank = xmalloc(*nbr_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  offset = 0;
  for (unsigned i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (unsigned i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (unsigned i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (unsigned i = 0; i < *nbr_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = xmalloc(part_num_cells * sizeof(**global_cell_id));
  *cell_core_mask = xmalloc(part_num_cells * sizeof(**cell_core_mask));
  *global_corner_id = xmalloc(part_num_vertices * sizeof(**global_corner_id));
  *corner_core_mask = xmalloc(part_num_vertices * sizeof(**corner_core_mask));

  *x_cells = xmalloc(part_num_cells * sizeof(**x_cells));
  *y_cells = xmalloc(part_num_cells * sizeof(**y_cells));
  *x_vertices = xmalloc(part_num_vertices * sizeof(**x_vertices));
  *y_vertices = xmalloc(part_num_vertices * sizeof(**y_vertices));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = xmalloc(*nbr_vertices * sizeof(*global_to_local_vertex));
  for (unsigned i = 0; i < *nbr_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*corner_core_mask)[part_num_vertices] = vertex_is_on_rank[i] == -1;
      double p[3] = {x_vertices_3d[i], y_vertices_3d[i], z_vertices_3d[i]};
      XYZtoLL(p, (*x_vertices) + part_num_vertices, (*y_vertices) + part_num_vertices);
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  free(vertex_is_on_rank);
  *nbr_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  offset = 0;
  for (unsigned i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*cell_core_mask)[part_num_cells] = (cell_is_on_rank[i] == -1);
      double middle_point[3] = {x_cell_3d[i],y_cell_3d[i],z_cell_3d[i]};

      

      for (unsigned j = 0; j < (*num_vertices_per_cell)[i]; ++j) {
        (*cell_to_vertex)[num_cell_vertex_dependencies+j] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];
      }

      XYZtoLL(middle_point, (*x_cells)+part_num_cells, (*y_cells)+part_num_cells);

      num_cell_vertex_dependencies += (*num_vertices_per_cell)[i];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  free(x_cell_3d);
  free(y_cell_3d);
  free(z_cell_3d);
  free(x_vertices_3d);
  free(y_vertices_3d);
  free(z_vertices_3d);

  *num_vertices_per_cell = xrealloc(*num_vertices_per_cell, part_num_cells *
                                    sizeof(**num_vertices_per_cell));
  *cell_to_vertex = xrealloc(*cell_to_vertex, num_cell_vertex_dependencies *
                             sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *nbr_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);
}

static struct grid_data generate_icon_grid_data(MPI_Comm comm) {

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  struct grid_data grid_data;

  static char const * icon_grid_name = "icon_grid";
  grid_data.name = icon_grid_name;

  read_part_icon_grid_information(
    "icon_grid_R02B02.nc", &grid_data.nbr_vertices, &grid_data.nbr_cells,
    &grid_data.num_vertices_per_cell, &grid_data.cell_to_vertex,
    &grid_data.x_vertices, &grid_data.y_vertices,
    &grid_data.x_cells, &grid_data.y_cells,
    &grid_data.global_cell_id, &grid_data.cell_mask,
    &grid_data.cell_core_mask, &grid_data.global_corner_id,
    &grid_data.corner_core_mask, comm_rank, comm_size);

  for (size_t i = 0; i < grid_data.nbr_cells; ++i)
    grid_data.cell_mask[i] = grid_data.cell_mask[i] == 0;

  return grid_data;
}

static struct grid_data generate_cube_grid_data(MPI_Comm comm) {

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  struct grid_data grid_data;

  static char const * cube_grid_name = "cube_grid";
  grid_data.name = cube_grid_name;

  unsigned n = 50;

  generate_part_cube_grid_information(
    n, (unsigned*)&grid_data.nbr_vertices, (unsigned*)&grid_data.nbr_cells,
    (unsigned**)&grid_data.num_vertices_per_cell,
    (unsigned**)&grid_data.cell_to_vertex,
    &grid_data.x_vertices, &grid_data.y_vertices,
    &grid_data.x_cells, &grid_data.y_cells, &grid_data.global_cell_id,
    &grid_data.cell_core_mask, &grid_data.global_corner_id,
    &grid_data.corner_core_mask, comm_rank, comm_size);

  grid_data.cell_mask = NULL;

  return grid_data;
}

static struct grid_data generate_mpiom_grid_data(MPI_Comm comm) {

  int comm_rank, comm_size;
  MPI_Comm_rank(comm, &comm_rank);
  MPI_Comm_size(comm, &comm_size);

  struct grid_data grid_data;

  static char const * mpiom_grid_name = "mpiom_grid";
  grid_data.name = mpiom_grid_name;

  read_part_mpiom_grid_information(
    "GR30_lsm.nc", &grid_data.nbr_vertices, &grid_data.nbr_cells,
    &grid_data.num_vertices_per_cell, &grid_data.cell_to_vertex,
    &grid_data.x_vertices, &grid_data.y_vertices,
    &grid_data.x_cells, &grid_data.y_cells,
    &grid_data.global_cell_id, &grid_data.cell_mask,
    &grid_data.cell_core_mask, &grid_data.global_corner_id,
    &grid_data.corner_core_mask, comm_rank, comm_size);

  for (size_t i = 0; i < grid_data.nbr_cells; ++i)
    grid_data.cell_mask[i] = grid_data.cell_mask[i] == 0;

  return grid_data;
}

static void grid_data_free(struct grid_data grid_data) {

  free(grid_data.num_vertices_per_cell);
  free(grid_data.cell_to_vertex);
  free(grid_data.x_vertices);
  free(grid_data.y_vertices);
  free(grid_data.x_cells);
  free(grid_data.y_cells);

  free(grid_data.cell_mask);
  free(grid_data.global_cell_id);
  free(grid_data.cell_core_mask);
  free(grid_data.global_corner_id);
  free(grid_data.corner_core_mask);
}
